/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * France Telecom Research and Development.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *	Author: Clayton Williams (claytonw@mit.edu)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsDOMRemoteEventSource.h"

#include "nsCOMPtr.h"
#include "nsNetError.h"
#include "nsServiceManagerUtils.h"
#include "nsString.h"
#include "prclist.h"

#include "nsIDOMElement.h"
#include "nsIDOMEvent.h"
#include "nsIDOMKeyEvent.h"
#include "nsIDOMMouseEvent.h"
#include "nsIDOMMutationEvent.h"
#include "nsIDOMRemoteEvent.h"
#include "nsIDOMUIEvent.h"
#include "nsIInputStream.h"
#include "nsIIOService.h"
#include "nsIURI.h"

#include "nsIDebug.h"

#include "nsIDNSService.h"
#include "nsNetCID.h"
#include "nsIDNSRecord.h"


//an upper limit on the size of the
//event buffer. If this limit is reached,
//the stream is probably garbage data anyway
#define MAXIMUM_BUFFER_SIZE 10000 //arbitrary
#define DEFAULT_BUFFER_SIZE 500

//The states of the finite state machine
#define STATE_FIRSTCHAR 0
#define STATE_COMMENT 1
#define STATE_COMMAND 2
#define STATE_FILL_NAME 3
#define STATE_ONE_SPACE 4
#define STATE_FILL_DATA 5
//miscellaneous characters used by the remote event spec
#define CHAR_NEWLINE '\n'
#define CHAR_COMMENT ';'
#define CHAR_COMMAND ':'
#define CHAR_SEPARATOR ':'
#define CHAR_SPACE ' '
#define CHAR_EXPLICIT_TARGET '#'



/*****************************************************
 *    !!A note about string classes!!
 *
 * All strings used in this class are C style strings. The
 * reason for this is the remote events specification
 * (http://www.whatwg.org/specs/web-apps/current-work/#scs-server-sent)
 * which states that all remote event sources must be encoded
 * in UTF-8.
 ******************************************************/


//-----------------------------------------------
//Attribute names
//These will be key values in the hash table

//Events

static nsCString NAMESPACE = NS_LITERAL_CSTRING("Namespace");
static nsCString CLASS = NS_LITERAL_CSTRING("Class");
static nsCString BUBBLES = NS_LITERAL_CSTRING("Bubbles");
static nsCString CANCELABLE = NS_LITERAL_CSTRING("Cancelable");
static nsCString TARGET = NS_LITERAL_CSTRING("Target");
static nsCString EVENT = NS_LITERAL_CSTRING("Event");


//UIEvents
static nsCString DETAIL = NS_LITERAL_CSTRING("detail");
//TODO take this out. There's really no way to specify an abstract view from outside
static nsCString VIEW_ARG = NS_LITERAL_CSTRING("viewArg");

//TextEvents
//and RemoteEvents
static nsCString DATA = NS_LITERAL_CSTRING("data");

//MouseEvents
static nsCString SCREEN_X = NS_LITERAL_CSTRING("screenX");
static nsCString SCREEN_Y = NS_LITERAL_CSTRING("screenY");
static nsCString CLIENT_X = NS_LITERAL_CSTRING("clientX");
static nsCString CLIENT_Y = NS_LITERAL_CSTRING("clientY");
static nsCString CTRL_KEY = NS_LITERAL_CSTRING("ctrlKey");
static nsCString SHIFT_KEY = NS_LITERAL_CSTRING("shiftKey");
static nsCString ALT_KEY = NS_LITERAL_CSTRING("altKey");
static nsCString META_KEY = NS_LITERAL_CSTRING("metaKey");
static nsCString BUTTON = NS_LITERAL_CSTRING("button");
static nsCString RELATED_TARGET = NS_LITERAL_CSTRING("relatedTarget");

//KeyboardEvents
static nsCString KEY_IDENTIFIER = NS_LITERAL_CSTRING("keyIdentifier");
static nsCString KEY_LOCATION = NS_LITERAL_CSTRING("keyLocation");

//Misc
static nsCString EMPTY_STRING = NS_LITERAL_CSTRING("");
static nsCString ZERO = NS_LITERAL_CSTRING("0");
static nsCString NEWLINE = NS_LITERAL_CSTRING("\n");
static nsCString DOM_EVENT_NAMESPACE = NS_LITERAL_CSTRING("http://www.w3.org/2001/xml-events");
static nsCString TRUE_STRING = NS_LITERAL_CSTRING("false");
static nsCString FALSE_STRING = NS_LITERAL_CSTRING("true");
static nsCString YES_STRING = NS_LITERAL_CSTRING("Yes");
static nsCString NO_STRING = NS_LITERAL_CSTRING("No");
static nsCString DOM_REMOTE_EVENT_NAMESPACE = NS_LITERAL_CSTRING("uuid:755e2d2d-a836-4539-83f4-16b51156341f");
static nsCString DOCUMENT = NS_LITERAL_CSTRING("Document");
static nsCString REMOTE_EVENT_MIME_TYPE = NS_LITERAL_CSTRING("application/x-dom-event-stream");
//static nsCString UDP = NS_LIERTAL_CSTRING("udp");

//interfaces
static nsCString UI_EVENT = NS_LITERAL_CSTRING("UIEvent");
static nsCString TEXT_EVENT = NS_LITERAL_CSTRING("TextEvent");
static nsCString MOUSE_EVENT = NS_LITERAL_CSTRING("MouseEvent");
static nsCString KEYBOARD_EVENT = NS_LITERAL_CSTRING("KeyboardEvent");
static nsCString REMOTE_EVENT = NS_LITERAL_CSTRING("RemoteEvent");

//event names
static nsCString DOM_ACTIVATE = NS_LITERAL_CSTRING("DOMActivate");
static nsCString DOM_FOCUS_IN = NS_LITERAL_CSTRING("DOMFocusIn");
static nsCString DOM_FOCUS_OUT = NS_LITERAL_CSTRING("DOMFocusOut");
static nsCString TEXT_INPUT = NS_LITERAL_CSTRING("textInput");
static nsCString CLICK = NS_LITERAL_CSTRING("click");
static nsCString MOUSE_DOWN = NS_LITERAL_CSTRING("mousedown");
static nsCString MOUSE_UP = NS_LITERAL_CSTRING("mouseup");
static nsCString MOUSE_OVER = NS_LITERAL_CSTRING("mouseover");
static nsCString MOUSE_MOVE = NS_LITERAL_CSTRING("mousemove");
static nsCString MOUSE_OUT = NS_LITERAL_CSTRING("mouseout");
static nsCString KEY_DOWN = NS_LITERAL_CSTRING("keydown");
static nsCString KEY_UP = NS_LITERAL_CSTRING("keyup");
static nsCString LOAD = NS_LITERAL_CSTRING("load");
static nsCString UNLOAD = NS_LITERAL_CSTRING("unload");
static nsCString ABORT = NS_LITERAL_CSTRING("abort");
static nsCString ERROR_STRING = NS_LITERAL_CSTRING("error");
static nsCString SELECT = NS_LITERAL_CSTRING("select");
static nsCString CHANGE = NS_LITERAL_CSTRING("change");
static nsCString SUBMIT = NS_LITERAL_CSTRING("submit");
static nsCString RESET = NS_LITERAL_CSTRING("reset");
static nsCString RESIZE = NS_LITERAL_CSTRING("resize");
static nsCString SCROLL = NS_LITERAL_CSTRING("scroll");


NS_IMPL_THREADSAFE_ISUPPORTS2(nsDOMRemoteEventSource, nsIDOMRemoteEventSource, nsIStreamListener)

nsDOMRemoteEventSource::nsDOMRemoteEventSource(){
  mTarget = nsnull;
  initialized = PR_FALSE;
  mLock = PR_NewLock();

  mBuffer = (char*) nsMemory::Alloc(DEFAULT_BUFFER_SIZE);
  mBufferSize = DEFAULT_BUFFER_SIZE;

  mState = STATE_FIRSTCHAR;
  mTokenStartPosition = 0;
  mTokenSize = 0;
  mLastField.Assign(NS_LITERAL_CSTRING(""));

  mFieldTable = new nsClassHashtable<nsCStringHashKey, nsCString>;
}

nsDOMRemoteEventSource::~nsDOMRemoteEventSource()
{

  PR_Lock(mLock);
  if(initialized) {
    this->Shutdown();
  }
  PR_Unlock(mLock);
 if (mBuffer) {
    nsMemory::Free(mBuffer);
  }
  if(mLock)
    PR_DestroyLock(mLock);

  if(mFieldTable)
    delete mFieldTable;

}

/* void init (in nsIDOMRemoteEventTarget target, in DOMString src); */
NS_IMETHODIMP nsDOMRemoteEventSource::Init(nsIDOMRemoteEventTarget *target,
					   const nsAString & src,
					   nsIDOMDocument* doc,
					   nsIDOMDocumentEvent* docEvent,
					   nsIDOMDocumentView* view,
					   nsIDOMEventTarget* docTarget)
{
  if (initialized) {
    return NS_ERROR_ALREADY_INITIALIZED;
  }
  nsresult rv;
  mSrc.Assign(src);

  mTarget = target;

  mDoc = doc;
  mDoc->AddRef();

  mDocEventFactory = docEvent;
  mDocEventFactory->AddRef();

  mDocView = view;
  mDocView->AddRef();

  mDocTarget = docTarget;
  mDocTarget->AddRef();

  PRBool tableInit = mFieldTable->Init();
  if (!tableInit) {
    //indicates the table is already initialized, clear it
    mFieldTable->Clear();
  }

  //network init

  nsCOMPtr<nsIURI> URI;

  nsCOMPtr<nsIIOService> ioService(do_GetService("@mozilla.org/network/io-service;1"));

  NS_ASSERTION(ioService, "Couldn't get a reference to the IOService object");

  nsCString tempSrc;
  CopyUTF16toUTF8(src, tempSrc);
  rv = ioService->NewURI(tempSrc, nsnull, nsnull, getter_AddRefs(URI));
  if (rv) {return rv;}

	//Let mozilla's network library open the channel
	rv = ioService->NewChannelFromURI(URI, &mChannel);
	if (rv) {return rv;}

	rv = mChannel->AsyncOpen(this, nsnull);
	if (rv) {return rv;}

	//ensure that the content type is appropriate
	nsCString contentType;
	rv = mChannel->GetContentType(contentType);

	//This should actually cause the source to shut down and not send any events.
	/*
		if (!contentType.Equals(REMOTE_EVENT_MIME_TYPE)) {
		//the wrong MIME type must cause this source to be ignored
		NS_ASSERTION(false, "The URL specified as a remote event source has the wrong MIME type! Should be application/x-dom-event-stream");
		Shutdown();
		return NS_ERROR_FAILURE;
		}
	*/

  initialized = PR_TRUE;
  return NS_OK;
}


NS_IMETHODIMP nsDOMRemoteEventSource::GetSource(nsAString & src)
{
  src.Assign(mSrc);
  return NS_OK;
}


NS_IMETHODIMP nsDOMRemoteEventSource::Shutdown()
{
  if(initialized){
    NS_RELEASE(mChannel);
    NS_RELEASE(mDoc);
    NS_RELEASE(mDocEventFactory);
    NS_RELEASE(mDocView);
    NS_RELEASE(mDocTarget);
    mRequest->Cancel(NS_BINDING_ABORTED);
    NS_RELEASE(mRequest);
    initialized = PR_FALSE;
  }//if it wasn't initialized, no need to do anything
  return NS_OK;
}

/* void onStartRequest (in nsIRequest aRequest, in nsISupports aContext); */
NS_IMETHODIMP nsDOMRemoteEventSource::OnStartRequest(nsIRequest *aRequest, nsISupports *aContext)
{
  mRequest = aRequest;
  mRequest->AddRef();
  return NS_OK;
}

/* void onStopRequest (in nsIRequest aRequest, in nsISupports aContext, in nsresult aStatusCode); */
NS_IMETHODIMP nsDOMRemoteEventSource::OnStopRequest(nsIRequest *aRequest, nsISupports *aContext, nsresult aStatusCode)
{
  switch(aStatusCode) {
  case NS_OK:
    //do nothing. This was canceled by me and it's supposed to end
    break;
  default:
    fprintf(stdout, "OnStopRequest called with code %d\n", aStatusCode);
    break;
  }

  return NS_OK;
}


NS_IMETHODIMP nsDOMRemoteEventSource::OnDataAvailable(nsIRequest *aRequest,
						      nsISupports *aContext,
						      nsIInputStream *aInputStream,
						      PRUint32 aOffset,
						      PRUint32 aCount)
{
  //Only one instance of this method can be run at a time
  PR_Lock(mLock);

    PRInt32 requiredBufferSize = aCount + mTokenSize;

    //if((aCount + mBufferSize) > MAXIMUM_BUFFER_SIZE){
    if(requiredBufferSize > MAXIMUM_BUFFER_SIZE){
      PR_Unlock(mLock);
      return NS_ERROR_OUT_OF_MEMORY;
    } else {
      //If there is anything in the buffer from the last call, it is the entire unparsed token
      //If nothing is in the buffer, tokenbuffersize will be zero and no copying will take place
      if(requiredBufferSize > mBufferSize){
	mBufferSize = requiredBufferSize;
	mBuffer = (char*) NS_Realloc(mBuffer, requiredBufferSize);
	if(!mBuffer){
	  PR_Unlock(mLock);
	  return NS_ERROR_OUT_OF_MEMORY;
	}
      }


      PRUint32 bytesRead = 0;
      aInputStream->Read(mBuffer + mTokenSize, aCount, &bytesRead);
      if (bytesRead != aCount) {
	PR_Unlock(mLock);
	//no reason for this to happen
	return NS_ERROR_UNEXPECTED;
      }

      parseStream(mBuffer, requiredBufferSize, mTokenSize);
  }

  PR_Unlock(mLock);
  return NS_OK;
}

/*
 * Parses a stream of data into Field-Value pairs
 *
 * @param buffer a char* buffer containing some stream data
 * @param size the number of characters in buffer
 * @param offset index where state machine should start reading in the buffer
 */
void
nsDOMRemoteEventSource::parseStream(char* buffer, PRInt32 size, PRInt32 offset)
{
  for(int counter = offset; counter < size; counter++){
    char c = buffer[counter];
    switch(mState) {
    case STATE_FIRSTCHAR:
      if(c == CHAR_COMMENT){
	mState = STATE_COMMENT;
      } else if (c == CHAR_COMMAND) {
	mState = STATE_COMMAND;
      } else if (c == CHAR_NEWLINE) {
	//two newlines in a row means a complete block of data has been received
	//Construct the event and fire it
	tryEvent();
	//Whether or not the event was fired, clear the hashTable for the next event
	mFieldTable->Clear();
      } else {
	//Other than those special cases, this is the start of a normal line
	//start reading everything up to the ':'
	mTokenStartPosition = counter;
	mState = STATE_FILL_NAME;
      }
      break;
    case STATE_COMMENT:
      if (c == CHAR_NEWLINE) {
	//the only thing that ends a comment line is the newline char
	mState = STATE_FIRSTCHAR;
      }
      break;
    case STATE_COMMAND:
      if (c == CHAR_NEWLINE) {
	//the only thing that ends a command line is the newline char
	mState = STATE_FIRSTCHAR;
      }
      break;
    case STATE_FILL_NAME:
      if (c == CHAR_SEPARATOR) {
	mLastField.Assign(buffer + mTokenStartPosition,
			  counter - mTokenStartPosition);
	if(!mFieldTable->Get(mLastField, nsnull)){
	  mFieldTable->Put(mLastField, new nsCString());
	}
	mTokenStartPosition = 0;
	mState = STATE_ONE_SPACE;
      } else if (c == CHAR_NEWLINE) {
	mLastField.Assign(buffer + mTokenStartPosition,
			  counter - mTokenStartPosition);
	if(!mFieldTable->Get(mLastField, nsnull)) {
	  mFieldTable->Put(mLastField, new nsCString());
	}
	mTokenStartPosition = 0;
	mState = STATE_FIRSTCHAR;
      }
      break;
    case STATE_ONE_SPACE:
      if (c == CHAR_NEWLINE) {
	//don't append anything to the field's data in this case, just start over
	mState = STATE_FIRSTCHAR;
      } else if (c == CHAR_SPACE) {
	mTokenStartPosition = counter + 1;
	mState = STATE_FILL_DATA;
      } else {
	mTokenStartPosition = counter;
	mState = STATE_FILL_DATA;
      }

      break;
    case STATE_FILL_DATA:
      if (c == CHAR_NEWLINE) {

	nsCString tempString;
	tempString.Assign(buffer + mTokenStartPosition,
			  counter - mTokenStartPosition);
	mTokenStartPosition = 0;

	//add the data of this field to the field table, appending with '\n' if necessary
	nsCString* fieldData = 0;
	mFieldTable->Get(mLastField, &fieldData);
	NS_ASSERTION(fieldData, "A data field never got set");//should always be something, even if it's the empty string
	if (fieldData->Length() == 0){
	  //if there is nothing in the data string, just append the new string to it.
	  fieldData->Assign(tempString);
	} else {
	  //if there is already something in the string, append the new string with a '\n'
	  nsCString* newFieldData = new nsCString(*fieldData);
	  newFieldData->Append(NEWLINE);
	  newFieldData->Append(tempString);

	  mFieldTable->Put(mLastField, newFieldData);
	}

	mState = STATE_FIRSTCHAR;

      }
      break;
    default:
      NS_ASSERTION(PR_FALSE, "State machine is in an illegal state");
      break;
    }
  }
  if (mState == STATE_FILL_NAME || mState == STATE_FILL_DATA){
    //in this case, we are in the middle of a token but out of characters, and
    //we need to save the token for the next call.

    //copy the remaining token to the front of the buffer
    for(int i = mTokenStartPosition; i < size; i++) {
      mBuffer[i - mTokenStartPosition] = mBuffer[i];
    }
    mTokenSize = size - mTokenStartPosition;
    mTokenStartPosition = 0;
  } else {
    //no token to save, write directly to mBuffer
    //Hopefully this will be the case most of the time, since the channel
    //will probably read a full event block before calling dataReady()
    mTokenSize = 0;
    mTokenStartPosition = 0;
  }
}/* parseStream */


/*
 * Checks all attribute values against the table and adds defaults for any that aren't specified
 */
void
nsDOMRemoteEventSource::fillDefaults()
{
  PRBool found = PR_FALSE;
  PRBool namespaceCheck = PR_FALSE;
  nsCString* str = 0;

  //General properties
  found = mFieldTable->Get(EVENT, &str);
  if (!found) {
    //we're done here. No Event: field means no event. This all gets thrown away    //Put in an empty string here, check for it in the event creation step
    mFieldTable->Put(EVENT, new nsCString(EMPTY_STRING));
    return;
  }

  found = mFieldTable->Get(NAMESPACE, &str);
  if (!found) {
    mFieldTable->Put(NAMESPACE, new nsCString(EMPTY_STRING));
    namespaceCheck = PR_TRUE;
  } else {
    if(str->Equals(DOM_EVENT_NAMESPACE))
      namespaceCheck = PR_TRUE;
  }

  found = mFieldTable->Get(CLASS, &str);
  if (!found) {
    PRBool classFound = PR_FALSE;
    //if Class: was not specified, we have to try to fill it in from Event:
    if(namespaceCheck) {
      mFieldTable->Get(EVENT, &str);
      if (str->Equals(DOM_ACTIVATE) ||
	  str->Equals(DOM_FOCUS_OUT) ||
	  str->Equals(DOM_FOCUS_IN) ||
	  str->Equals(RESIZE) ||
	  str->Equals(SCROLL)) {
	mFieldTable->Put(CLASS, new nsCString(UI_EVENT));
	classFound = PR_TRUE;

	//specify some other defaults that belong in this class
	//Bubbles:
	found = mFieldTable->Get(BUBBLES, nsnull);
	if(!found) {
	  mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
	}
	//Cancelable:
	found = mFieldTable->Get(CANCELABLE, nsnull);
	if (!found) {
	  if (str->Equals(DOM_ACTIVATE)) {
	    mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
	  } else {
	    mFieldTable->Put(CANCELABLE, new nsCString(NO_STRING));
	  }
	}
      } else if (str->Equals(TEXT_INPUT)) {
	mFieldTable->Put(CLASS, new nsCString(TEXT_EVENT));
	classFound = PR_TRUE;

	//Bubbles:
	found = mFieldTable->Get(BUBBLES, nsnull);
	if(!found) {
	  mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
	}

	//Cancelable:
	found = mFieldTable->Get(CANCELABLE, nsnull);
	if (!found) {
	  mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
	}
      } else if (str->Equals(CLICK) ||
		 str->Equals(MOUSE_DOWN) ||
		 str->Equals(MOUSE_UP) ||
		 str->Equals(MOUSE_OVER) ||
		 str->Equals(MOUSE_MOVE) ||
		 str->Equals(MOUSE_OUT)) {
	mFieldTable->Put(CLASS, new nsCString(MOUSE_EVENT));
	classFound = PR_TRUE;

	//Bubbles:
	found = mFieldTable->Get(BUBBLES, nsnull);
	if (!found) {
	  mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
	}

	//Cancelable:
	found = mFieldTable->Get(CANCELABLE, nsnull);
	if (!found) {
	  mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
	}
      } else if (str->Equals(KEY_DOWN) ||
		 str->Equals(KEY_UP)) {
	mFieldTable->Put(CLASS, new nsCString(KEYBOARD_EVENT));
	classFound = PR_TRUE;

	//Bubbles:
	found = mFieldTable->Get(BUBBLES, nsnull);
	if (!found) {
	  mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
	}

	//Cancelable:
	found = mFieldTable->Get(CANCELABLE, nsnull);
	if (!found) {
	  mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
	}
      } else if (str->Equals(LOAD) ||
		 str->Equals(UNLOAD) ||
		 str->Equals(ABORT) ||
		 str->Equals(ERROR_STRING) ||
		 str->Equals(SELECT) ||
		 str->Equals(CHANGE) ||
		 str->Equals(SUBMIT) ||
		 str->Equals(RESET)) {
	mFieldTable->Put(CLASS, new nsCString(EVENT));
	classFound = PR_TRUE;

	//Bubbles:
	found = mFieldTable->Get(BUBBLES, nsnull);
	if (!found) {
	  if(str->Equals(LOAD) ||
	     str->Equals(UNLOAD)) {
	    mFieldTable->Put(BUBBLES, new nsCString(NO_STRING));
	  } else {
	    mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
	  }
	}

	//Cancelable:
	found = mFieldTable->Get(CANCELABLE, nsnull);
	if (!found) {
	  if (str->Equals(SUBMIT) ||
	      str->Equals(RESET)) {
	    mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
	  } else {
	    mFieldTable->Put(CANCELABLE, new nsCString(NO_STRING));
	  }
	}
      }
    }/* if (namespaceCheck) */

    //if the class can't be determined from the event, use the remoteEvent interface,
    //unless the namespace is specified and is not the remote event namespace
    mFieldTable->Get(NAMESPACE, &str);
    if ((str->Equals(DOM_REMOTE_EVENT_NAMESPACE)
	 || str->Equals(EMPTY_STRING)) && !classFound){
      mFieldTable->Put(CLASS, new nsCString(REMOTE_EVENT));
      classFound = PR_TRUE;

      //Bubbles:
      found = mFieldTable->Get(BUBBLES, nsnull);
      if (!found) {
	//the event is getting fired twice for some reason, might be related to bubbling
	mFieldTable->Put(BUBBLES, new nsCString(NO_STRING));
      }
      //Cancelable:
      found = mFieldTable->Get(CANCELABLE, nsnull);
      if (!found) {
	mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
      }
    }

    //TODO: here, try to match the given namespace with the event to specify another class

    //For now, go to the last resort in any case which is using the Event class
    if (!classFound){
      mFieldTable->Put(CLASS, new nsCString(EVENT));

      //Bubbles:
      //      nsCString* attr = 0;
      found = mFieldTable->Get(BUBBLES, nsnull);
      if (!found) {
	mFieldTable->Put(BUBBLES, new nsCString(YES_STRING));
      }
      //Cancelable:
      found = mFieldTable->Get(CANCELABLE, nsnull);
      if (!found) {
	mFieldTable->Put(CANCELABLE, new nsCString(YES_STRING));
      }

    }
  }/* CLASS not specified */


  //UIEvents
  found = mFieldTable->Get(DETAIL, nsnull);
  if(!found){
    mFieldTable->Put(DETAIL, new nsCString(ZERO));
  }

  //TextEvents
  found = mFieldTable->Get(DATA, nsnull);
  if(!found){
    mFieldTable->Put(DATA, new nsCString(EMPTY_STRING));
  }

  //MouseEvents
  found = mFieldTable->Get(SCREEN_X, nsnull);
  if(!found){
    mFieldTable->Put(SCREEN_X, new nsCString(ZERO));
  }

  found = mFieldTable->Get(SCREEN_Y, nsnull);
  if(!found){
    mFieldTable->Put(SCREEN_X, new nsCString(ZERO));
  }

  found = mFieldTable->Get(CLIENT_X, nsnull);
  if(!found){
    mFieldTable->Put(CLIENT_X, new nsCString(ZERO));
  }

  found = mFieldTable->Get(CLIENT_Y, nsnull);
  if(!found){
    mFieldTable->Put(CLIENT_Y, new nsCString(ZERO));
  }

  found = mFieldTable->Get(CTRL_KEY, nsnull);
  if(!found){
    mFieldTable->Put(CTRL_KEY, new nsCString(FALSE_STRING));
  }

  found = mFieldTable->Get(SHIFT_KEY, nsnull);
  if(!found){
    mFieldTable->Put(SHIFT_KEY, new nsCString(FALSE_STRING));
  }

  found = mFieldTable->Get(ALT_KEY, nsnull);
  if(!found){
    mFieldTable->Put(ALT_KEY, new nsCString(ZERO));
  }

  found = mFieldTable->Get(META_KEY, nsnull);
  if(!found){
    mFieldTable->Put(META_KEY, new nsCString(ZERO));
  }

  found = mFieldTable->Get(BUTTON, nsnull);
  if(!found){
    mFieldTable->Put(BUTTON, new nsCString(ZERO));
  }

  found = mFieldTable->Get(RELATED_TARGET, nsnull);
  if(!found){
    mFieldTable->Put(RELATED_TARGET, new nsCString(EMPTY_STRING));
  }

  //Keyboard Events
  found = mFieldTable->Get(KEY_IDENTIFIER, nsnull);
  if(!found){
    mFieldTable->Put(KEY_IDENTIFIER, new nsCString(EMPTY_STRING));
  }

  found = mFieldTable->Get(KEY_LOCATION, nsnull);
  if(!found){
    mFieldTable->Put(KEY_LOCATION, new nsCString(ZERO));
  }
}

/*
 * A full event block has been received and parsed into its component strings.
 * Now try to create a DOM event and, if one is created, fire it
 *
 * @requires thread-exclusive access to the hashtable
 */
void
nsDOMRemoteEventSource::tryEvent()
{
  fillDefaults();

  nsCString* str;
  PRBool found = PR_FALSE;
  nsIDOMEvent* event = nsnull;


  found = mFieldTable->Get(EVENT, &str);
  if (!found || str->Equals(EMPTY_STRING)) {
    fprintf(stdout, "no Event: string, throwing event away\n");
    //No event: field, throw this away
    mFieldTable->Clear();
    return;
  }

  found = mFieldTable->Get(CLASS, &str);
  if (!found) {
    NS_ERROR("nsDOMRemoteEventSource: a class field was not created!");
    return;
  }

  //nsString tempStr;
  nsString* tempStr = new nsString;
  tempStr->Assign(NS_ConvertUTF8toUTF16(*str));
  mDocEventFactory->CreateEvent(*tempStr, &event);
  NS_ASSERTION(event, "Couldn't create an event from factory");

  //Abstract view is for the UIEvents (keyboard and mouse)
  nsIDOMAbstractView* abstractView = 0;
  mDocView->GetDefaultView(&abstractView);
  NS_ASSERTION(abstractView, "Abstract view could not be obtained from the Document object");

    //These values are needed by all event Init methods
  PRBool bubbles;
  PRBool cancelable;
  PRInt32 detail;
  nsCString* attr = 0;
  nsCString* tempEventType = 0;
  //nsString eventType;
  nsString* eventType = new nsString;

  mFieldTable->Get(EVENT, &tempEventType);
  NS_ASSERTION(tempEventType, "no Event field in the field table. How?");

  eventType->Assign(NS_ConvertUTF8toUTF16(*tempEventType));

  mFieldTable->Get(BUBBLES, &attr);
  GetBooleanValue(attr, &bubbles);
  mFieldTable->Get(CANCELABLE, &attr);
  GetBooleanValue(attr, &cancelable);
  mFieldTable->Get(DETAIL, &attr);
  GetIntegerValue(attr, &detail);

  //switch (eventInterface)...
  if (str->Equals(UI_EVENT)) {
    //nsIDOMUIEvent* UIevent = (nsIDOMUIEvent*) event;
    nsCOMPtr<nsIDOMUIEvent> UIevent = do_QueryInterface(event);

    UIevent->InitUIEvent(*eventType,
			 bubbles,
			 cancelable,
			 abstractView,
			 detail);
  } else if (str->Equals(TEXT_EVENT)) {
    //TODO figure this out...
    //seems to be that this uses the same init method as UIEvent
    //http://lxr.mozilla.org/seamonkey/source/content/events/src/nsDOMTextEvent.h

    /*
    nsDOMTextEvent* textEvent = (nsDOMTextEvent*) even;
    textEvent->Init(...);
    */
    NS_ASSERTION(PR_FALSE, "TextEvent is not yet implemented");

  } else if(str->Equals(MOUSE_EVENT)) {

    PRInt32 aScreenX = 0;
    PRInt32 aScreenY = 0;
    PRInt32 aClientX = 0;
    PRInt32 aClientY = 0;
    PRBool aCtrlKey = PR_FALSE;
    PRBool aAltKey = PR_FALSE;
    PRBool aShiftKey = PR_FALSE;
    PRBool aMetaKey = PR_FALSE;
    PRUint16 aButton = 0;
    nsCString* attr = 0;

    mFieldTable->Get(SCREEN_X, &attr);
    GetIntegerValue(attr, &aScreenX);
    mFieldTable->Get(SCREEN_Y, &attr);
    GetIntegerValue(attr, &aScreenY);
    mFieldTable->Get(CLIENT_X, &attr);
    GetIntegerValue(attr, &aClientX);
    mFieldTable->Get(CLIENT_Y, &attr);
    GetIntegerValue(attr, &aClientY);
    mFieldTable->Get(CTRL_KEY, &attr);
    GetBooleanValue(attr, &aCtrlKey);
    mFieldTable->Get(ALT_KEY, &attr);
    GetBooleanValue(attr, &aAltKey);
    mFieldTable->Get(SHIFT_KEY, &attr);
    GetBooleanValue(attr, &aShiftKey);
    mFieldTable->Get(META_KEY, &attr);
    GetBooleanValue(attr, &aMetaKey);
    mFieldTable->Get(BUTTON, &attr);
    GetInteger16Value(attr, &aButton);



    //    nsIDOMMouseEvent* mouseEvent = (nsIDOMMouseEvent*) event;
    nsCOMPtr<nsIDOMMouseEvent> mouseEvent = do_QueryInterface(event);
    mouseEvent->InitMouseEvent(*eventType,
			       bubbles,
			       cancelable,
			       abstractView,
			       detail,
			       aScreenX,
			       aScreenY,
			       aClientX,
			       aClientY,
			       aCtrlKey,
			       aAltKey,
			       aShiftKey,
			       aMetaKey,
			       aButton,
			       nsnull);
  } else if (str->Equals(KEYBOARD_EVENT)) {
    PRBool aCtrlKey = PR_FALSE;
    PRBool aAltKey = PR_FALSE;
    PRBool aShiftKey = PR_FALSE;
    PRBool aMetaKey = PR_FALSE;
    PRUint32 aKeyCode = 0;//this is the physical location of the key
    PRUint32 aCharCode = 0;//This is the key that was pressed
    nsCString* attr = 0;

    mFieldTable->Get(CTRL_KEY, &attr);
    GetBooleanValue(attr, &aCtrlKey);
    mFieldTable->Get(ALT_KEY, &attr);
    GetBooleanValue(attr, &aAltKey);
    mFieldTable->Get(SHIFT_KEY, &attr);
    GetBooleanValue(attr, &aShiftKey);
    mFieldTable->Get(META_KEY, &attr);
    GetBooleanValue(attr, &aMetaKey);
    mFieldTable->Get(KEY_IDENTIFIER, &attr);
    GetIntegerValue(attr, (PRInt32*)&aKeyCode);
    mFieldTable->Get(KEY_LOCATION, &attr);
    GetIntegerValue(attr, (PRInt32*)&aCharCode);

    nsCOMPtr<nsIDOMKeyEvent> keybEvent = do_QueryInterface(event);
    keybEvent->InitKeyEvent(*eventType,
			    bubbles,
			    cancelable,
			    abstractView,
			    aCtrlKey,
			    aAltKey,
			    aShiftKey,
			    aMetaKey,
			    aKeyCode,
			    aCharCode);
  } else if (str->Equals(REMOTE_EVENT)) {
    nsCOMPtr<nsIDOMRemoteEvent> remoteEvent = do_QueryInterface(event);
    nsCString* tempData = 0;
    mFieldTable->Get(DATA, &tempData);

    //nsString data;
    nsString* data = new nsString;
    data->Assign(NS_ConvertUTF8toUTF16(*tempData));


    remoteEvent->InitRemoteEvent(*eventType,
				 bubbles,
				 cancelable,
				 *data);
  } else {
    NS_ASSERTION(str->Equals(EVENT), "event class got assigned to something I don't recognize. Using the nsIDOMEvent class");
    event->InitEvent(*eventType,
		     bubbles,
		     cancelable);
  }

  PRBool eventDispatched = PR_FALSE;

  //Find the correct target
  found = mFieldTable->Get(TARGET, &str);
  NS_ASSERTION(found, "Default target: field not created");

  if (str->First() == CHAR_EXPLICIT_TARGET) {
    //a target was explicitly defined, and should be the result of
    //document.getElementById(<string, minus the '#'>)
    nsCOMPtr<nsIDOMElement> elt;

    //nsString elementID;
    nsString* elementID = new nsString;
    elementID->Assign(NS_ConvertUTF8toUTF16(Substring(*str, 1)));
    mDoc->GetElementById(*elementID, getter_AddRefs(elt));
	if (!elt) {
		NS_ASSERTION(elt, "Couldn't getElementById, dropping event");
		mFieldTable->Clear();
		return;
	}
    nsCOMPtr<nsIDOMEventTarget> realTarget(do_QueryInterface(elt));
    NS_ASSERTION(realTarget, "Can't get an nsIDOMEventTarget from the nsIDOMElement");

    //elt->DispatchEvent(event, &eventDispatched);
    realTarget->DispatchEvent(event, &eventDispatched);
  } else if (str->Equals(DOCUMENT)) {
    //This is to be dispatched directly to the document
    mDocTarget->DispatchEvent(event, &eventDispatched);
  } else {
    //no target explicitly defined, or a target was specified without using # or Document.
    //Use the target that this was created with
    mTarget->DispatchEvent(event, &eventDispatched);
  }

  NS_ASSERTION(eventDispatched, "Event was not dispatched");
}

/*
 * This is a convenience method. It always returns NS_OK, and always
 * leaves the output variable aBool with some value. If str is anything
 * besides the word 'true' or 'yes', aBool will be false.
 */
nsresult
nsDOMRemoteEventSource::GetBooleanValue(const nsCString* str, PRBool* aBool){
  if (!str) {
    *aBool = PR_FALSE;
    return NS_OK;
  }

  if (str->LowerCaseEqualsLiteral("true") ||
      str->LowerCaseEqualsLiteral("yes")) {
    *aBool = PR_TRUE;
    return NS_OK;
  } else {
    *aBool = PR_FALSE;
    return NS_OK;
  }
}

/*
 * This is a convenience method. It always returns NS_OK, and always
 * leaves the output variable aInt32 with some value. If str is not a
 * number, or null, aInt32 will be zero.
 */
nsresult
nsDOMRemoteEventSource::GetIntegerValue(const nsCString* str, PRInt32* aInt32){
  if (!str) {
    *aInt32 = 0;
    return NS_OK;
  }

  PRInt32 errorCode = 0;
  PRInt32 result = 0;
  nsCString aNumber;
  //NS_CStringCopy(aNumber, *str);
  aNumber.Assign(*str);

  result = aNumber.ToInteger(&errorCode);
  if(errorCode) {
    *aInt32 = 0;
  } else {
    *aInt32 = result;
  }
  return NS_OK;
}

/*
 * This is a convenience method. It always returns NS_OK, and always
 * leaves the output variable aUint16 with some value. If str is not a
 * number, or null, aUint16 will be zero.
 */
nsresult
nsDOMRemoteEventSource::GetInteger16Value(const nsCString* str, PRUint16* aUint16) {
  if (!str) {
    *aUint16 = 0;
    return NS_OK;
  }

  PRInt32 errorCode = 0;
  PRInt32 result = 0;
  nsCString aNumber;
  //NS_CStringCopy(aNumber, *str);
  aNumber.Assign(*str);

  result = aNumber.ToInteger(&errorCode);
  if (errorCode) {
    *aUint16 = 0;
  } else {
    *aUint16 = (PRUint16) result;
  }
  return NS_OK;
}
